@@ -28,6 +28,7 @@
 #endif
 
 #include <math.h>
+#include <float.h>
 #include <linux/input.h>
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -55,6 +56,8 @@
 
 #include "clutter-device-manager-evdev.h"
 
+#define DISCRETE_SCROLL_STEP 10.0
+
 #define AUTOREPEAT_VALUE 2
 
 /* Try to keep the pointer inside the stage. Hopefully no one is using
@@ -100,6 +103,10 @@ struct _ClutterSeatEvdev
 
   gfloat pointer_x;
   gfloat pointer_y;
+
+  /* Emulation of discrete scroll events out of smooth ones */
+  gfloat accum_scroll_dx;
+  gfloat accum_scroll_dy;
 };
 
 struct _ClutterEventFilter
@@ -137,9 +144,14 @@ struct _ClutterDeviceManagerEvdevPrivate
   GSList *event_filters;
 };
 
-G_DEFINE_TYPE_WITH_PRIVATE (ClutterDeviceManagerEvdev,
-                            clutter_device_manager_evdev,
-                            CLUTTER_TYPE_DEVICE_MANAGER)
+static void clutter_device_manager_evdev_event_extender_init (ClutterEventExtenderInterface *iface);
+
+G_DEFINE_TYPE_WITH_CODE (ClutterDeviceManagerEvdev,
+                         clutter_device_manager_evdev,
+                         CLUTTER_TYPE_DEVICE_MANAGER,
+                         G_ADD_PRIVATE (ClutterDeviceManagerEvdev)
+                         G_IMPLEMENT_INTERFACE (CLUTTER_TYPE_EVENT_EXTENDER,
+                                                clutter_device_manager_evdev_event_extender_init))
 
 static ClutterOpenDeviceCallback  device_open_callback;
 static ClutterCloseDeviceCallback device_close_callback;
@@ -171,6 +183,36 @@ static const char *option_xkb_layout = "us";
 static const char *option_xkb_variant = "";
 static const char *option_xkb_options = "";
 
+static void
+clutter_device_manager_evdev_copy_event_data (ClutterEventExtender *event_extender,
+                                              const ClutterEvent   *src,
+                                              ClutterEvent         *dest)
+{
+  ClutterEventEvdev *event_evdev;
+
+  event_evdev = _clutter_event_get_platform_data (src);
+  if (event_evdev != NULL)
+    _clutter_event_set_platform_data (dest, _clutter_event_evdev_copy (event_evdev));
+}
+
+static void
+clutter_device_manager_evdev_free_event_data (ClutterEventExtender *event_extender,
+                                              ClutterEvent         *event)
+{
+  ClutterEventEvdev *event_evdev;
+
+  event_evdev = _clutter_event_get_platform_data (event);
+  if (event_evdev != NULL)
+    _clutter_event_evdev_free (event_evdev);
+}
+
+static void
+clutter_device_manager_evdev_event_extender_init (ClutterEventExtenderInterface *iface)
+{
+  iface->copy_event_data = clutter_device_manager_evdev_copy_event_data;
+  iface->free_event_data = clutter_device_manager_evdev_free_event_data;
+}
+
 /*
  * ClutterEventSource for reading input devices
  */
@@ -258,7 +300,7 @@ notify_key_device (ClutterInputDevice *input_device,
   /* We can drop the event on the floor if no stage has been
    * associated with the device yet. */
   stage = _clutter_input_device_get_stage (input_device);
-  if (!stage)
+  if (stage == NULL)
     {
       clear_repeat_timer (seat);
       return;
@@ -270,6 +312,7 @@ notify_key_device (ClutterInputDevice *input_device,
 					     seat->xkb,
 					     seat->button_state,
 					     time_, key, state);
+  _clutter_evdev_event_set_event_code (event, key);
 
   /* We must be careful and not pass multiple releases to xkb, otherwise it gets
      confused and locks the modifiers */
@@ -360,12 +403,7 @@ notify_absolute_motion (ClutterInputDevice *input_device,
   ClutterStage *stage;
   ClutterEvent *event = NULL;
 
-  /* We can drop the event on the floor if no stage has been
-   * associated with the device yet. */
   stage = _clutter_input_device_get_stage (input_device);
-  if (!stage)
-    return;
-
   device_evdev = CLUTTER_INPUT_DEVICE_EVDEV (input_device);
   manager_evdev = CLUTTER_DEVICE_MANAGER_EVDEV (input_device->device_manager);
   seat = _clutter_input_device_evdev_get_seat (device_evdev);
@@ -428,11 +466,70 @@ notify_relative_motion (ClutterInputDevice *input_device,
   notify_absolute_motion (input_device, time_, new_x, new_y);
 }
 
+static ClutterScrollDirection
+discrete_to_direction (gdouble discrete_x,
+                       gdouble discrete_y)
+{
+  if (discrete_x > 0)
+    return CLUTTER_SCROLL_RIGHT;
+  else if (discrete_x < 0)
+    return CLUTTER_SCROLL_LEFT;
+  else if (discrete_y > 0)
+    return CLUTTER_SCROLL_DOWN;
+  else if (discrete_y < 0)
+    return CLUTTER_SCROLL_UP;
+  else
+    return CLUTTER_SCROLL_SMOOTH;
+}
+
+static void
+notify_discrete_scroll (ClutterInputDevice     *input_device,
+                        guint32                 time_,
+                        ClutterScrollDirection  direction,
+                        gboolean                emulated)
+{
+  ClutterInputDeviceEvdev *device_evdev;
+  ClutterSeatEvdev *seat;
+  ClutterStage *stage;
+  ClutterEvent *event = NULL;
+
+  if (direction == CLUTTER_SCROLL_SMOOTH)
+    return;
+
+  /* We can drop the event on the floor if no stage has been
+   * associated with the device yet. */
+  stage = _clutter_input_device_get_stage (input_device);
+  if (stage == NULL)
+    return;
+
+  device_evdev = CLUTTER_INPUT_DEVICE_EVDEV (input_device);
+  seat = _clutter_input_device_evdev_get_seat (device_evdev);
+
+  event = clutter_event_new (CLUTTER_SCROLL);
+
+  event->scroll.time = time_;
+  event->scroll.stage = CLUTTER_STAGE (stage);
+  event->scroll.device = seat->core_pointer;
+  _clutter_xkb_translate_state (event, seat->xkb, seat->button_state);
+
+  event->scroll.direction = direction;
+
+  event->scroll.x = seat->pointer_x;
+  event->scroll.y = seat->pointer_y;
+  clutter_event_set_device (event, seat->core_pointer);
+  clutter_event_set_source_device (event, input_device);
+
+  _clutter_event_set_pointer_emulated (event, emulated);
+
+  queue_event (event);
+}
+
 static void
 notify_scroll (ClutterInputDevice *input_device,
                guint32             time_,
                gdouble             dx,
-               gdouble             dy)
+               gdouble             dy,
+               gboolean            emulated)
 {
   ClutterInputDeviceEvdev *device_evdev;
   ClutterSeatEvdev *seat;
@@ -443,7 +540,7 @@ notify_scroll (ClutterInputDevice *input_device,
   /* We can drop the event on the floor if no stage has been
    * associated with the device yet. */
   stage = _clutter_input_device_get_stage (input_device);
-  if (!stage)
+  if (stage == NULL)
     return;
 
   device_evdev = CLUTTER_INPUT_DEVICE_EVDEV (input_device);
@@ -460,7 +557,7 @@ notify_scroll (ClutterInputDevice *input_device,
    * To convert to Xi2 discrete step coordinate space, multiply the factor
    * 1/10. */
   event->scroll.direction = CLUTTER_SCROLL_SMOOTH;
-  scroll_factor = 1.0 / 10.0;
+  scroll_factor = 1.0 / DISCRETE_SCROLL_STEP;
   clutter_event_set_scroll_delta (event,
                                   scroll_factor * dx,
                                   scroll_factor * dy);
@@ -470,6 +567,8 @@ notify_scroll (ClutterInputDevice *input_device,
   clutter_event_set_device (event, seat->core_pointer);
   clutter_event_set_source_device (event, input_device);
 
+  _clutter_event_set_pointer_emulated (event, emulated);
+
   queue_event (event);
 }
 
@@ -493,7 +592,7 @@ notify_button (ClutterInputDevice *input_device,
   /* We can drop the event on the floor if no stage has been
    * associated with the device yet. */
   stage = _clutter_input_device_get_stage (input_device);
-  if (!stage)
+  if (stage == NULL)
     return;
 
   device_evdev = CLUTTER_INPUT_DEVICE_EVDEV (input_device);
@@ -517,11 +616,12 @@ notify_button (ClutterInputDevice *input_device,
       break;
 
     default:
-      button_nr = button - BTN_MOUSE + 1;
+      /* For compatibility reasons, all additional buttons go after the old 4-7 scroll ones */
+      button_nr = button - (BTN_LEFT - 1) + 4;
       break;
     }
 
-  if (G_UNLIKELY (button_nr < 1 || button_nr > 8))
+  if (button_nr < 1 || button_nr > 12)
     {
       g_warning ("Unhandled button event 0x%x", button);
       return;
@@ -548,6 +648,8 @@ notify_button (ClutterInputDevice *input_device,
   clutter_event_set_device (event, seat->core_pointer);
   clutter_event_set_source_device (event, input_device);
 
+  _clutter_evdev_event_set_event_code (event, button);
+
   _clutter_input_device_set_stage (seat->core_pointer, stage);
 
   queue_event (event);
@@ -569,7 +671,7 @@ notify_touch_event (ClutterInputDevice *input_device,
   /* We can drop the event on the floor if no stage has been
    * associated with the device yet. */
   stage = _clutter_input_device_get_stage (input_device);
-  if (!stage)
+  if (stage == NULL)
     return;
 
   device_evdev = CLUTTER_INPUT_DEVICE_EVDEV (input_device);
@@ -614,7 +716,7 @@ notify_pinch_gesture_event (ClutterInputDevice          *input_device,
   /* We can drop the event on the floor if no stage has been
    * associated with the device yet. */
   stage = _clutter_input_device_get_stage (input_device);
-  if (!stage)
+  if (stage == NULL)
     return;
 
   device_evdev = CLUTTER_INPUT_DEVICE_EVDEV (input_device);
@@ -659,7 +761,7 @@ notify_swipe_gesture_event (ClutterInputDevice          *input_device,
   /* We can drop the event on the floor if no stage has been
    * associated with the device yet. */
   stage = _clutter_input_device_get_stage (input_device);
-  if (!stage)
+  if (stage == NULL)
     return;
 
   device_evdev = CLUTTER_INPUT_DEVICE_EVDEV (input_device);
@@ -1198,6 +1300,39 @@ _device_seat_get_touch (ClutterInputDevice *input_device,
   return g_hash_table_lookup (seat->touches, GUINT_TO_POINTER (id));
 }
 
+static void
+check_notify_discrete_scroll (ClutterDeviceManagerEvdev *manager_evdev,
+                              ClutterInputDevice        *device,
+                              guint32                    time_)
+{
+  ClutterInputDeviceEvdev *device_evdev =
+    CLUTTER_INPUT_DEVICE_EVDEV (device);
+  ClutterSeatEvdev *seat = _clutter_input_device_evdev_get_seat (device_evdev);
+  int i, n_xscrolls, n_yscrolls;
+
+  n_xscrolls = floor (fabs (seat->accum_scroll_dx) / DISCRETE_SCROLL_STEP);
+  n_yscrolls = floor (fabs (seat->accum_scroll_dy) / DISCRETE_SCROLL_STEP);
+
+  for (i = 0; i < n_xscrolls; i++)
+    {
+      notify_discrete_scroll (device, time_,
+                              seat->accum_scroll_dx > 0 ?
+                              CLUTTER_SCROLL_RIGHT : CLUTTER_SCROLL_LEFT,
+                              TRUE);
+    }
+
+  for (i = 0; i < n_yscrolls; i++)
+    {
+      notify_discrete_scroll (device, time_,
+                              seat->accum_scroll_dy > 0 ?
+                              CLUTTER_SCROLL_DOWN : CLUTTER_SCROLL_UP,
+                              TRUE);
+    }
+
+  seat->accum_scroll_dx = fmodf (seat->accum_scroll_dx, DISCRETE_SCROLL_STEP);
+  seat->accum_scroll_dy = fmodf (seat->accum_scroll_dy, DISCRETE_SCROLL_STEP);
+}
+
 static gboolean
 process_device_event (ClutterDeviceManagerEvdev *manager_evdev,
                       struct libinput_event *event)
@@ -1261,7 +1396,7 @@ process_device_event (ClutterDeviceManagerEvdev *manager_evdev,
         device = libinput_device_get_user_data (libinput_device);
 
         stage = _clutter_input_device_get_stage (device);
-        if (!stage)
+        if (stage == NULL)
           break;
 
         stage_width = clutter_actor_get_width (CLUTTER_ACTOR (stage));
@@ -1306,14 +1441,17 @@ process_device_event (ClutterDeviceManagerEvdev *manager_evdev,
     case LIBINPUT_EVENT_POINTER_AXIS:
       {
         gdouble dx = 0.0, dy = 0.0;
+        gdouble discrete_x = 0.0, discrete_y = 0.0;
         guint32 time;
         gboolean wheel = FALSE;
         enum libinput_pointer_axis axis;
         enum libinput_pointer_axis_source source;
         struct libinput_event_pointer *axis_event =
           libinput_event_get_pointer_event (event);
+        ClutterSeatEvdev *seat;
 
         device = libinput_device_get_user_data (libinput_device);
+        seat = _clutter_input_device_evdev_get_seat (CLUTTER_INPUT_DEVICE_EVDEV (device));
 
         time = libinput_event_pointer_get_time (axis_event);
         source = libinput_event_pointer_get_axis_source (axis_event);
@@ -1329,22 +1467,41 @@ process_device_event (ClutterDeviceManagerEvdev *manager_evdev,
         axis = LIBINPUT_POINTER_AXIS_SCROLL_VERTICAL;
         if (libinput_event_pointer_has_axis (axis_event, axis))
           {
-            if (wheel)
-              dy = 10 * libinput_event_pointer_get_axis_value_discrete (axis_event, axis);
+            discrete_y = libinput_event_pointer_get_axis_value_discrete (axis_event, axis);
+            dy = libinput_event_pointer_get_axis_value (axis_event, axis);
+
+            if (wheel || fabs (dy) < DBL_EPSILON)
+              seat->accum_scroll_dy = 0;
             else
-              dy = libinput_event_pointer_get_axis_value (axis_event, axis);
+              seat->accum_scroll_dy += dy;
           }
 
         axis = LIBINPUT_POINTER_AXIS_SCROLL_HORIZONTAL;
         if (libinput_event_pointer_has_axis (axis_event, axis))
           {
-            if (wheel)
-              dx = 10 * libinput_event_pointer_get_axis_value_discrete (axis_event, axis);
+            discrete_x = libinput_event_pointer_get_axis_value_discrete (axis_event, axis);
+            dx = libinput_event_pointer_get_axis_value (axis_event, axis);
+
+            if (wheel || fabs (dx) < DBL_EPSILON)
+              seat->accum_scroll_dx = 0;
             else
-              dx = libinput_event_pointer_get_axis_value (axis_event, axis);
+              seat->accum_scroll_dx += dx;
+          }
+
+        if (wheel)
+          {
+            notify_scroll (device, time,
+                           discrete_x * DISCRETE_SCROLL_STEP,
+                           discrete_y * DISCRETE_SCROLL_STEP,
+                           TRUE);
+            notify_discrete_scroll (device, time, discrete_to_direction (discrete_x, discrete_y), FALSE);
+          }
+        else
+          {
+            notify_scroll (device, time, dx, dy, FALSE);
+            check_notify_discrete_scroll (manager_evdev, device, time);
           }
 
-        notify_scroll (device, time, dx, dy);
         break;
 
       }
@@ -1362,7 +1519,7 @@ process_device_event (ClutterDeviceManagerEvdev *manager_evdev,
         device = libinput_device_get_user_data (libinput_device);
 
         stage = _clutter_input_device_get_stage (device);
-        if (!stage)
+        if (stage == NULL)
           break;
 
         stage_width = clutter_actor_get_width (CLUTTER_ACTOR (stage));
@@ -1417,7 +1574,7 @@ process_device_event (ClutterDeviceManagerEvdev *manager_evdev,
         device = libinput_device_get_user_data (libinput_device);
 
         stage = _clutter_input_device_get_stage (device);
-        if (!stage)
+        if (stage == NULL)
           break;
 
         stage_width = clutter_actor_get_width (CLUTTER_ACTOR (stage));
